@@ -15,12 +15,14 @@ from glue.core.message import SubsetDeleteMessage, SubsetUpdateMessage
 from glue.core.decorators import memoize
 from glue.core.visual import VisualAttributes
 from glue.config import settings
-from glue.utils import view_shape
+from glue.utils import view_shape, broadcast_to
+
 
 __all__ = ['Subset', 'SubsetState', 'RoiSubsetState', 'CategoricalROISubsetState',
            'RangeSubsetState', 'MultiRangeSubsetState', 'CompositeSubsetState',
            'OrState', 'AndState', 'XorState', 'InvertState', 'MaskSubsetState', 'CategorySubsetState',
-           'ElementSubsetState', 'InequalitySubsetState', 'combine_multiple']
+           'ElementSubsetState', 'InequalitySubsetState', 'combine_multiple',
+           'CategoricalMultiRangeSubsetState', 'CategoricalROISubsetState2D']
 
 
 OPSYM = {operator.ge: '>=', operator.gt: '>',
@@ -166,9 +168,12 @@ class Subset(object):
         return np.where(self._to_mask_join(None).flat)[0]
 
     def _to_mask_join(self, view):
-        """Conver the subset to a mask through an entity join
-           to another dataset. """
+        """
+        Convert the subset to a mask through an entity join to another
+        dataset.
+        """
         for other, (cid1, cid2) in self.data._key_joins.items():
+
             if getattr(other, '_recursing', False):
                 continue
 
@@ -176,17 +181,66 @@ class Subset(object):
                 self.data._recursing = True
                 s2 = Subset(other)
                 s2.subset_state = self.subset_state
-                key_right = s2.to_mask()
+                mask_right = s2.to_mask()
             except IncompatibleAttribute:
                 continue
             finally:
                 self.data._recursing = False
 
-            key_left = self.data[cid1, view]
-            result = np.in1d(key_left.ravel(),
-                             other[cid2, key_right])
+            if len(cid1) == 1 and len(cid2) == 1:
 
-            return result.reshape(key_left.shape)
+                key_left = self.data[cid1[0], view]
+                key_right = other[cid2[0], mask_right]
+                mask = np.in1d(key_left.ravel(), key_right.ravel())
+
+                return mask.reshape(key_left.shape)
+
+            elif len(cid1) == len(cid2):
+
+                key_left_all = []
+                key_right_all = []
+
+                for cid1_i, cid2_i in zip(cid1, cid2):
+                    key_left_all.append(self.data[cid1_i, view].ravel())
+                    key_right_all.append(other[cid2_i, mask_right].ravel())
+
+                # TODO: The following is slow because we are looping in Python.
+                #       This could be made significantly faster by switching to
+                #       C/Cython.
+
+                key_left_all = zip(*key_left_all)
+                key_right_all = set(zip(*key_right_all))
+
+                result = [key in key_right_all for key in key_left_all]
+                result = np.array(result)
+
+                return result.reshape(self.data[cid1_i, view].shape)
+
+            elif len(cid1) == 1:
+
+                key_left = self.data[cid1[0], view].ravel()
+                mask = np.zeros_like(key_left, dtype=bool)
+                for cid2_i in cid2:
+                    key_right = other[cid2_i, mask_right].ravel()
+                    mask |= np.in1d(key_left, key_right)
+
+                return mask.reshape(self.data[cid1[0], view].shape)
+
+            elif len(cid2) == 1:
+
+                key_right = other[cid2[0], mask_right].ravel()
+                mask = np.zeros_like(self.data[cid1[0], view].ravel(), dtype=bool)
+                for cid1_i in cid1:
+                    key_left = self.data[cid1_i, view].ravel()
+                    mask |= np.in1d(key_left, key_right)
+
+                return mask.reshape(self.data[cid1[0], view].shape)
+
+            else:
+
+                raise Exception("Either the number of components in the key join sets "
+                                "should match, or one of the component sets should ",
+                                "contain a single component.")
 
         raise IncompatibleAttribute
 
@@ -433,13 +487,58 @@ class RoiSubsetState(SubsetState):
     def attributes(self):
         return (self.xatt, self.yatt)
 
-    @memoize
     @contract(data='isinstance(Data)', view='array_view')
     def to_mask(self, data, view=None):
+
+        # TODO: make sure that pixel components don't actually take up much
+        #       memory and are just views
+
         x = data[self.xatt, view]
         y = data[self.yatt, view]
-        result = self.roi.contains(x, y)
-        assert x.shape == result.shape
+
+        # We do the import here to avoid circular imports
+        from glue.core.component_id import PixelComponentID
+
+        if (x.ndim == data.ndim and
+            isinstance(self.xatt, PixelComponentID) and
+            isinstance(self.yatt, PixelComponentID)):
+
+            # This is a special case - the ROI is defined in pixel space, so we
+            # can apply it to a single slice and then broadcast it to all other
+            # dimensions. We start off by extracting a slice which takes only
+            # the first elements of all dimensions except the attributes in
+            # question, for which we take all the elements. We need to preserve
+            # the dimensionality of the array, hence the use of slice(0, 1).
+            # Note that we can only do this if the view (if present) preserved
+            # the dimensionality, which is why we checked that x.ndim == data.ndim
+
+            subset = []
+            for i in range(data.ndim):
+                if i == self.xatt.axis or i == self.yatt.axis:
+                    subset.append(slice(None))
+                else:
+                    subset.append(slice(0, 1))
+
+            x_slice = x[subset]
+            y_slice = y[subset]
+
+            if self.roi.defined():
+                result = self.roi.contains(x_slice, y_slice)
+            else:
+                result = np.zeros(x_slice.shape, dtype=bool)
+
+            result = broadcast_to(result, x.shape)
+
+        else:
+
+            if self.roi.defined():
+                result = self.roi.contains(x, y)
+            else:
+                result = np.zeros(x.shape, dtype=bool)
+
+        if result.shape != x.shape:
+            raise ValueError("Unexpected error: boolean mask has incorrect dimensions")
+
         return result
 
     def copy(self):
@@ -537,6 +636,124 @@ class MultiRangeSubsetState(SubsetState):
         return MultiRangeSubsetState(self.pairs, self.att)
 
 
+class CategoricalROISubsetState2D(object):
+    """
+    A 2D subset state where both attributes are categorical.
+
+    Parameters
+    ----------
+    categories : dict
+        A dictionary containing for each label of one categorical component an
+        interable of labels for the other categorical component (using sets will
+        provide the best performance)
+    att1 : :class:`~glue.core.component_id.ComponentID`
+        The component ID matching the keys of the ``categories`` dictionary
+    att2 : :class:`~glue.core.component_id.ComponentID`
+        The component ID matching the values of the ``categories`` dictionary
+    """
+    def __init__(self, categories, att1, att2):
+        self.categories = categories
+        self.att1 = att1
+        self.att2 = att2
+
+    @property
+    def attributes(self):
+        return (self.att1, self.att2)
+
+    @memoize
+    @contract(data='isinstance(Data)', view='array_view')
+    def to_mask(self, data, view=None):
+
+        # Extract categories and numerical values
+        labels1 = data.get_component(self.att1).labels
+        labels2 = data.get_component(self.att2).labels
+
+        if view is not None:
+            labels1 = labels1[view]
+            labels2 = labels2[view]
+
+        # Initialize empty mask
+        mask = np.zeros(labels1.shape, dtype=bool)
+
+        # A loop over all values here is actually reasonably efficient compared
+        # to alternatives. Any improved implementation, even vectorized, should
+        # ensure that it is more efficient for large numbers of categories and
+        # values.
+        for i in range(len(labels1)):
+            if labels1[i] in self.categories:
+                if labels2[i] in self.categories[labels1[i]]:
+                    mask[i] = True
+
+        return mask
+
+    def copy(self):
+        result = CategoricalROISubsetState2D(self.categories,
+                                             self.att1, self.att2)
+        return result
+
+
+class CategoricalMultiRangeSubsetState(SubsetState):
+    """
+    A 2D subset state where one attribute is categorical and the other is
+    numerical, and where for each category, there are multiple possible subset
+    ranges.
+
+    Parameters
+    ----------
+    ranges : dict
+        A dictionary containing for each category (key), a list of tuples
+        giving the ranges of values for the numerical attribute.
+    cat_att : :class:`~glue.core.component_id.ComponentID`
+        The component ID for the categorical attribute
+    num_att : :class:`~glue.core.component_id.ComponentID`
+        The component ID for the numerical attribute
+    """
+
+    def __init__(self, ranges, cat_att, num_att):
+        self.ranges = ranges
+        self.cat_att = cat_att
+        self.num_att = num_att
+
+    @property
+    def attributes(self):
+        return (self.cat_att, self._num_att)
+
+    @memoize
+    @contract(data='isinstance(Data)', view='array_view')
+    def to_mask(self, data, view=None):
+
+        # Extract categories and numerical values
+        labels = data.get_component(self.cat_att).labels
+        values = data[self.num_att]
+
+        if view is not None:
+            labels = labels[view]
+            values = values[view]
+
+        # Initialize empty mask
+        mask = np.zeros(values.shape, dtype=bool)
+
+        # A loop over all values here is actually reasonably efficient compared
+        # to alternatives. Any improved implementation, even vectorized, should
+        # ensure that it is more efficient for large numbers of categories and
+        # values. For example, using 10000 categories and 1000000 data points
+        # takes 1.2 seconds on a laptop.
+        for i in range(len(values)):
+            if labels[i] in self.ranges:
+                for lo, hi in self.ranges[labels[i]]:
+                    if values[i] >= lo and values[i] <= hi:
+                        mask[i] = True
+                        break
+
+        return mask
+
+    def copy(self):
+        result = CategoricalMultiRangeSubsetState(self.ranges,
+                                                  self.cat_att,
+                                                  self.num_att)
+        return result
+
+
 class CompositeSubsetState(SubsetState):
     op = None
 
@@ -567,7 +784,6 @@ class CompositeSubsetState(SubsetState):
         sym = OPSYM.get(self.op, self.op)
         return "(%s %s %s)" % (self.state1, sym, self.state2)
 
-
 class OrState(CompositeSubsetState):
     op = operator.or_
 
@@ -605,7 +821,11 @@ class MaskSubsetState(SubsetState):
         self.cids = cids
         self.mask = mask
 
+    def copy(self):
+        return MaskSubsetState(self.mask, self.cids)
+
     def to_mask(self, data, view=None):
+
         view = view or slice(None)
 
         # shortcut for data on the same pixel grid
@@ -659,29 +879,41 @@ class CategorySubsetState(SubsetState):
 
 class ElementSubsetState(SubsetState):
 
-    def __init__(self, indices=None):
+    def __init__(self, indices=None, data=None):
         super(ElementSubsetState, self).__init__()
         self._indices = indices
+        if data is None:
+            self._data_uuid = None
+        else:
+            self._data_uuid = data.uuid
 
     @memoize
     def to_mask(self, data, view=None):
-        # XXX this is inefficient for views
-        result = np.zeros(data.shape, dtype=bool)
-        if self._indices is not None:
-            result.flat[self._indices] = True
-        if view is not None:
-            result = result[view]
-        return result
+        if data.uuid == self._data_uuid or self._data_uuid is None:
+            # XXX this is inefficient for views
+            result = np.zeros(data.shape, dtype=bool)
+            if self._indices is not None:
+                result.flat[self._indices] = True
+            if view is not None:
+                result = result[view]
+            return result
+        else:
+            raise IncompatibleAttribute()
 
     def copy(self):
-        return ElementSubsetState(self._indices)
+        state = ElementSubsetState(indices=self._indices)
+        state._data_uuid = self._data_uuid
+        return state
 
     def __gluestate__(self, context):
-        return dict(indices=context.do(self._indices))
+        return dict(indices=context.do(self._indices),
+                    data_uuid=self._data_uuid)
 
     @classmethod
     def __setgluestate__(cls, rec, context):
-        return cls(indices=context.object(rec['indices']))
+        state = cls(indices=context.object(rec['indices']))
+        state._data_uuid = rec['data_uuid']
+        return state
 
 
 class InequalitySubsetState(SubsetState):
